import base64
import json
from functools import reduce

from aws_lambda_powertools.utilities.data_classes import (
    S3Event,
    APIGatewayProxyEvent,
    KafkaEvent,
    EventBridgeEvent
)
from aws_lambda_powertools.utilities.data_classes.s3_event import S3EventRecord
from aws_lambda_powertools.utilities.data_classes.kafka_event import KafkaEventRecord
from metis_fn import fn

from . import app_value, app_route, app_web_session

DEFAULT_S3_BUCKET_SEP = "."
NO_MATCHING_ROUTE = "no_matching_route"


def event_factory(event: dict,
                  factory_overrides: dict = None,
                  event_source_class=None) -> app_value.RequestEvent:
    if event_source_class:
        ev_fn, ev = event_match_fn_from_event_source(event, event_source_class)
    else:
        ev_fn, ev = dynamic_event_match(event)
    return ev_fn(ev, factory_overrides if factory_overrides else {})


def dynamic_event_match(event):
    ev_keys = event.keys()
    if 'Records' in ev_keys:
        return build_s3_state_change_event, S3Event(event)
    if 'httpMethod' in ev_keys:
        return build_http_event, APIGatewayProxyEvent(event)
    if 'eventSource' in ev_keys:
        return build_kafka_event, KafkaEvent(event)
    if 'source' in ev_keys:
        return build_event_bridge_event, EventBridgeEvent(event)
    return build_noop_event, event


def event_match_fn_from_event_source(event, event_source_cls):
    match event_source_cls.__name__:
        case 'S3Event':
            return build_s3_state_change_event, event_source_cls(event)
        case _:
            return build_noop_event, event


def build_noop_event(event: app_value.RequestEvent, _factory_overrides: dict = None) -> app_value.RequestEvent:
    template, route_fn, opts = route_fn_from_kind(NO_MATCHING_ROUTE)
    return app_value.NoopEvent(event=event,
                               kind=template,
                               request_function=route_fn)


def build_s3_state_change_event(event: S3Event, factory_overrides: dict) -> app_value.S3StateChangeEvent:
    objects = _s3_objects_from_event(event)
    factory = _domain_from_bucket_name if not factory_overrides.get('s3', None) else factory_overrides.get('s3', None)
    kind = factory(objects)
    template, route_fn, _opts = route_fn_from_kind(kind)
    return app_value.S3StateChangeEvent(event=event,
                                        kind=kind,
                                        request_function=route_fn,
                                        objects=objects)


def build_kafka_event(event: KafkaEvent, factory_overrides) -> app_value.KafkaRecordsEvent:
    evs = _kafka_events_from_event(event)
    factory = _domain_from_kafka_topic if not factory_overrides.get('kafka', None) else factory_overrides.get('kafka',
                                                                                                              None)
    kind = factory(evs)
    template, route_fn, _opts = route_fn_from_kind(kind)
    return app_value.KafkaRecordsEvent(event=event,
                                       kind=kind,
                                       request_function=route_fn,
                                       events=evs)


def build_event_bridge_event(event: EventBridgeEvent, factory_overrides) -> app_value.EventBridgePublishEvent:
    if not factory_overrides.get('eventbridge', None):
        factory = _domain_from_event_bridge_topic
    else:
        factory_overrides.get('eventbridge', None)

    kind = factory(event.detail_type)
    template, route_fn, _opts = route_fn_from_kind(kind)
    return app_value.EventBridgePublishEvent(topic=event.detail_type,
                                             kind=kind,
                                             event=event,
                                             body=event.detail,
                                             request_function=route_fn)


def build_http_event(event: APIGatewayProxyEvent,
                     _factory_overrides: dict) -> app_value.ApiGatewayRequestEvent:
    """
    method: str
    headers: dict
    resource: str
    body: str
    query_params: Optional[dict]=None
    """
    kind = _route_from_http_event(event.http_method, event.path)
    template, route_fn, opts = route_fn_from_kind(kind)
    hdrs = _standardise_headers(event.headers)
    body = _body_parser(opts, event)
    return app_value.ApiGatewayRequestEvent(kind=kind,
                                            request_function=route_fn,
                                            event=event,
                                            method=event['httpMethod'],
                                            headers=hdrs,
                                            path=event['path'],
                                            path_params=_path_template_to_params(kind[2], template[2]),
                                            body=body,
                                            query_params=event['queryStringParameters'],
                                            web_session=app_web_session.WebSession().session_from_headers(hdrs))


def _body_parser(route_options, event: APIGatewayProxyEvent):
    body = event.body
    if route_options and (parse_fn := route_options.get('body_parser', None)):
        return parse_fn(body)
    if event.is_base64_encoded:
        return base64.b64decode(body).decode('utf-8')
    return body


def route_fn_from_kind(kind):
    """
    Assumes that noop_event function is defined
    """
    return app_route.RouteMap().get_route(kind)


def _s3_objects_from_event(s3_event: S3Event) -> list[dict]:
    return [_s3_object(s3_event.bucket_name, record) for record in s3_event.records]


def _kafka_events_from_event(event: KafkaEvent) -> list[dict]:
    """
    Assumes that the event is in JSON
    """
    return [_kafka_event(record) for record in event.records]


def _domain_from_bucket_name(objects: list[app_value.S3Object]) -> str:
    """
    The handler is expected to deal with only 1 bucket as its a domain concept.  Multiple buckets
    indicate concern separation issues.
    """
    domain = {obj.bucket for obj in objects}
    if len(domain) > 1:
        return NO_MATCHING_ROUTE
    return domain.pop().split(DEFAULT_S3_BUCKET_SEP)[0]


def _domain_from_kafka_topic(events: list[KafkaEventRecord]) -> str:
    """
    The handler is expected to deal with only 1 topic as it's a domain concept.  Multiple buckets
    indicate concern separation issues.
    """
    domain = {ev.topic for ev in events}
    if len(domain) > 1:
        return NO_MATCHING_ROUTE
    return domain.pop()


def _domain_from_event_bridge_topic(topic) -> str:
    if not topic:
        return NO_MATCHING_ROUTE
    return topic


def _s3_object(bucket_name, record: S3EventRecord) -> app_value.S3Object:
    return app_value.S3Object(bucket=bucket_name,
                              key=record.s3.get_object.key)


def _kafka_event(record: KafkaEventRecord) -> app_value.KafkaTopicEvent:
    return app_value.KafkaTopicEvent(topic=record.topic,
                                     key=record.decoded_key if 'key' in record._data.keys() else None,
                                     value=record.json_value)


def _route_from_http_event(method, path):
    return ('API', method, path)


def _path_template_to_params(kind, template) -> dict:
    """
    Remove the leading "/"
    """
    return _params_comparer_builder(kind[1::].split("/"), template[1::].split("/"), {})


def _params_comparer_builder(kind_xs, template_xs, injector):
    template_fst, template_rst = fn.first(template_xs), fn.rest(template_xs)
    kind_fst, kind_rst = fn.first(kind_xs), fn.rest(kind_xs)
    if not template_fst:
        return injector
    if "{" in template_fst:
        return _params_comparer_builder(kind_rst,
                                        template_rst,
                                        {**injector, **{template_fst.replace("{", "").replace("}", ""): kind_fst}})
    return _params_comparer_builder(kind_rst, template_rst, injector)


def _standardise_headers(hdrs):
    def to_lower(acc, kv):
        k, v = kv
        acc[k.lower()] = v
        return acc

    return reduce(to_lower, hdrs.items(), {})



from typing import List, Dict
import uuid

from . import env


class SpanTracer():

    def __init__(self,
                 environment: env.EnvironmentProtocol,
                 tags: list[str] = None,
                 kv: dict[str, str] = None,
                 span_id: str = None):
        self.environment = environment
        self.span_id = span_id if span_id else str(uuid.uuid4())
        self.tags = tags if tags else []
        self.kv = kv if kv else {}
        self.trace_id = uuid.uuid4()

    def span_child(self, tags: list[str] = [], kv: dict[str, str] = {}):
        child = SpanTracer(environment=self.environment,
                           span_id=self.span_id,
                           tags=tags,
                           kv=kv)
        return child

    def serialise(self):
        return {**{'env': self.environment.env,
                   'trace_id': self.uuid_to_s(self.trace_id),
                   'span_id': self.span_id,
                   'tags': self.tags}, **self.kv}

    def uuid_to_s(self, uu_id):
        return str(uu_id) if isinstance(uu_id, uuid.UUID) else uu_id

    def aws_request_id(self):
        return self.kv.get('handler_id', None)


def init_tracing(env):
    return SpanTracer(env)



import json
import logging
from typing import Any, Protocol

from metis_fn import singleton
from pino import pino
import time

from .tracer import Tracer
from . import json_util

class ConfiguredLoggerProtocol(Protocol):

    def info(self, meta: dict, msg: str, **kwargs):
        ...

    def warn(self, meta: dict, msg: str, **kwargs):
        ...

    def error(self, meta: dict, msg: str, **kwargs):
        ...

    def debug(self, meta: dict, msg: str, **kwargs):
        ...


class LogConfig(singleton.Singleton):
    """
    Configure the logger and log level for all logging.  When using the log functions in this module (info, error, etc)
    the logger configured in LogConfig will be used as the logger.  The default logger is the Pino logger.
    The configured logger must implement the ConfiguredLoggerProtocol
    Note also that when using powertools via the observable module, the configured logged is the Powertools Logger cls.
    """
    default_level: int = logging.INFO
    configured_logger: Any = None

    def clear(self):
        self.configured_logger = None
        if getattr(self, 'logging_level', None):
            self.logging_level = None
        return self

    def configure(self, level: str | int = None, custom_logger: Any = None):
        if level:  # don't override the current level
            if isinstance(level, str):
                self.logging_level = self._level_from_name(level)
            else:
                self.logging_level = level
        self.configured_logger = custom_logger if custom_logger else self._standard_logger(self.level)
        return self

    def _level_from_name(self, name):
        if (lvl := getattr(logging, name.upper())):
            return lvl
        return

    @property
    def logger(self):
        if not getattr(self, 'configured_logger', None):
            return self._standard_logger(self.level)
        return self.configured_logger

    @property
    def level(self):
        if not getattr(self, 'logging_level', None):
            return self.default_level
        return self.logging_level

    def _standard_logger(self, level):
        return pino(bindings={"apptype": "prototype", "context": "main"},
                    dump_function=custom_pino_dump_fn,
                    level=self._level_as_name(level))

    def _level_as_name(self, lvl):
        return logging.getLevelName(lvl).lower()


class PowerToolsLoggerWrapper(ConfiguredLoggerProtocol):
    """
    The powertools logger is required to be wrapped to support this modules log function signature, which is
    consistent with the Pino logger (that is, an optional first arg of a context dict, and the second arg as a
    msg str.  For the Powertools logger, the meta (the ctx + other state) is passes as kwargs.  Implying that the
    meta dict will not be logged as a 'ctx' key in the structured logs, rather will be at the top level.
    """

    def __init__(self, lgr):
        self.logger = lgr

    def info(self, meta, msg):
        self.logger.info(msg, **meta)

    def warning(self, meta, msg):
        self.logger.warning(msg, **meta)

    def error(self, meta, msg):
        self.logger.error(msg, **meta)

    def debug(self, meta, msg):
        self.logger.debug(msg, **meta)


def info(msg: str,
         ctx: dict | None = None,
         tracer: Tracer | None = None,
         **kwargs) -> None:
    _log('info', msg, tracer, ctx if ctx else {}, **kwargs)


def debug(msg: str,
          ctx: dict | None = None,
          tracer: Tracer | None = None,
          **kwargs) -> None:
    _log('debug', msg, tracer, ctx if ctx else {}, **kwargs)


def warn(msg: str,
         ctx: dict | None = None,
         tracer: Tracer | None = None,
         **kwargs) -> None:
    _log('warning', msg, tracer, ctx if ctx else {}, **kwargs)


def error(msg: str,
          ctx: dict | None = None,
          tracer: Tracer | None = None,
          **kwargs) -> None:
    _log('error', msg, tracer, ctx if ctx else {}, **kwargs)


def _log(level: str,
         msg: str,
         tracer: Any,
         ctx: dict[str, str],
         **kwargs) -> None:
    if level not in level_functions.keys():
        return
    level_functions.get(level, info)(logger(), msg, meta(tracer, ctx, **kwargs))


def with_perf_log(perf_log_type: str = None, name: str = None):
    """
    Decorator which wraps the fn in a timer and writes a performance log
    """

    def inner(fn):
        def invoke(*args, **kwargs):
            t1 = time.time()
            result = fn(*args, **kwargs)
            t2 = time.time()
            if perf_log_type == 'http' and 'name' in kwargs:
                fn_name = kwargs['name']
            else:
                fn_name = name or fn.__name__
            perf_log(fn=fn_name, delta_t=(t2 - t1) * 1000.0)
            return result

        return invoke

    return inner


def log_decorator(fn):
    def log_writer(*args, **kwargs):
        _log(
            level='info',
            msg='Handling Command {fn}'.format(fn=fn.__name__),
            ctx=args[0].event,
            tracer=args[0].tracer
        )
        return fn(*args, **kwargs)

    return log_writer


def custom_pino_dump_fn(json_log):
    return json.dumps(json_log, cls=json_util.CustomLogEncoder)


def logger():
    return LogConfig().logger


def _info(lgr, msg: str, meta: dict) -> None:
    lgr.info(meta, msg)


def _debug(lgr, msg: str, meta: dict) -> None:
    lgr.debug(meta, msg)


def _warn(lgr, msg: str, meta: dict) -> None:
    lgr.warn(meta, msg)


def _error(lgr, msg: str, meta: dict) -> None:
    lgr.error(meta, msg)


def perf_log(fn: str, delta_t: float, callback: callable = None):
    if callback:
        callback(fn, delta_t)
    info("PerfLog", fn=fn, delta_t=delta_t)


def meta(tracer, ctx: dict, **kwargs):
    _meta = {**trace_meta(tracer),
             **kwargs}
    if not ctx:
        return _meta
    return {**_meta, **ctx}


def trace_meta(tracer):
    return tracer.serialise() if tracer else {}


level_functions = {'info': _info, 'error': _error, 'warn': _warn, 'debug': _debug}



import uuid

class Tracer():

    def __init__(self, env, aws_context=None):
        self.env = env
        self.aws_context = aws_context
        self.handler_id = uuid.uuid4()

    def serialise(self):
        return {'env': self.env, 'handler_id': self.handler_id_to_s(), 'aws_request_id': self.aws_request_id()}

    def handler_id_to_s(self):
        return str(self.handler_id) if isinstance(self.handler_id, uuid.UUID) else self.handler_id

    def aws_request_id(self):
        return self.aws_context.aws_request_id if self.aws_context else None


def init_tracing(env):
    return Tracer(env)



class BaseError(Exception):

    def __init__(self,
                 message="",
                 name="",
                 ctx={},
                 request_kwargs: dict = {},
                 traceback=None,
                 code=500,
                 klass="",
                 exception_klass=None,
                 retryable=False):
        self.code = 500 if code is None else code
        self.retryable = retryable
        self.message = message
        self.name = name
        self.ctx = ctx
        self.klass = klass
        self.exception_klass = exception_klass
        self.traceback = traceback
        self.request_kwargs = request_kwargs
        super().__init__(self.message)

    def error(self):
        return {'error': self.message, 'code': self.code, 'step': self.name, 'ctx': self.ctx}

    def print(self):
        print(f"{self.message}\n\n{self.traceback}")


class ParameterStoreError(BaseError):
    ...



from typing import Optional, List, Dict, Tuple, Callable, Any, Protocol
import json

from metis_fn import monad


class SerialiserProtocol(Protocol):
    def __init__(self, serialisable: Any, serialisation: Callable):
        ...

    def serialise(self):
        ...

    @property
    def content_type(self):
        ...


class DictToJsonSerialiser(SerialiserProtocol):
    CONTENT_TYPE = "application/json"

    def __init__(self, serialisable, serialisaton=None):
        self.serialisable = serialisable
        self.serialisation = serialisaton

    def serialise(self):
        return json.dumps(self.serialisable)

    @property
    def content_type(self):
        return self.__class__.CONTENT_TYPE


class DictToJsonLDSerialiser(DictToJsonSerialiser):
    CONTENT_TYPE = "application/ld+json"


def json_parser(body: str) -> str:
    """
    Attempts to parse the body as JSON.  If it fails it just returns the body
    """
    parsed = try_parser(json.loads, body)
    if parsed.is_right():
        return parsed.value
    return body


@monad.monadic_try()
def try_parser(parser_fn, content):
    return parser_fn(content)


def noon_parser(body: str) -> str:
    return body



import json
import decimal
import pendulum


class CustomLogEncoder(json.JSONEncoder):
    def default(self, obj):
        match type(obj):
            case pendulum.DateTime:
                return obj.to_iso8601_string()
            case decimal.Decimal:
                return str(obj)
            case _:
                return json.JSONEncoder.default(self, obj)


import sys
from typing import Any, Callable, Dict, Optional
from simple_memory_cache import GLOBAL_CACHE
from dataclasses import dataclass, field
from functools import reduce

from metis_fn import singleton

"""
Configures an AWS Client context object which initialises various AWS Clients.  The context is cached so that it can be 
used anywhere in the app.

Creating aws client instances is typically performed outside the main lambda handler, allowing the state to be cached between
lambda invocations.

First initialise the aws client config with the services you want and the boto3 lib you're using;
> services = {'s3': {}, 'ssm': {},  'dynamodb': {'table': 'table1'}}
> AwsClientConfig().configure(region_name="ap_southeast_2", aws_client_lib=boto3, services=services)

Then get the context when required:
>aws_ctx().ssm
  
"""

aws_cache = GLOBAL_CACHE.MemoryCachedVar('aws_cache')


@dataclass
class AwsCtx():
    s3: Optional[Any] = None
    ssm: Optional[Any] = None
    event_bridge: Optional[Any] = None
    table: Optional[Any] = None
    cognito_idp: Optional[Any] = None


class AwsClientConfig(singleton.Singleton):

    def configure(self,
                  aws_client_lib: Callable,
                  region_name: str,
                  services: Dict):
        self.aws_client_lib = aws_client_lib
        self.region_name = region_name
        self.services = services
        pass


def aws_ctx():
    return aws_cache.get()


def invalidate_cache():
    aws_cache.invalidate()
    pass


@aws_cache.on_first_access
def initiate_ctx() -> AwsCtx:
    ctx = reduce(client_builder, AwsClientConfig().services.items(), {})
    return AwsCtx(**ctx)


def client_builder(injector, service):
    return {**injector, **getattr(sys.modules[__name__], service[0])(service[1])}


def s3(args):
    return {'s3': AwsClientConfig().aws_client_lib.client('s3', region_name=AwsClientConfig().region_name)}


def ssm(args):
    return {'ssm': AwsClientConfig().aws_client_lib.client('ssm', region_name=AwsClientConfig().region_name)}


def events(args):
    return {
        'event_bridge': AwsClientConfig().aws_client_lib.client('events', region_name=AwsClientConfig().region_name)}


def dynamodb(args):
    return {'table': AwsClientConfig().aws_client_lib.resource('dynamodb', region_name=AwsClientConfig().region_name)
    .Table(args['table'])}


def cognito_idp(args):
    return {'cognito_idp': AwsClientConfig().aws_client_lib.client('cognito-idp',
                                                                   region_name=AwsClientConfig().region_name)}


def result_status(try_result) -> str:
    statuses = {1: 'ok'}
    if try_result.is_right():
        return statuses.get(try_result.lift()['ConsumedCapacity']['CapacityUnits'], 'fail')
    else:
        return 'fail'



from typing import List, Tuple, Callable
from dataclasses import dataclass
from pymonad.reader import Pipe
from pymonad.tools import curry

from metis_fn import fn, monad

@dataclass
class StateTransitionMap:
    stmap: List[Tuple[str, str, str]]

def state_transition_map(stmap: List[Tuple[str]]) -> StateTransitionMap:
    return StateTransitionMap(stmap)

def valid_transitions_from(state_map: StateTransitionMap, from_state: str) -> List[str]:
    return (Pipe(state_map)
           .then(map_selector(extract_from_state, (from_state, None, None)))
           .then(extract_transition_from_map)
           .flush())

def valid_state_transition(state_map: StateTransitionMap, from_state: str, with_transition: str) -> bool:
    return (Pipe(state_map)
           .then(map_selector(extract_from_state_and_transition, (from_state, with_transition, None)))
           .then(list)
           .then(fn.only_one)
           .flush())

def transition(state_map: StateTransitionMap, from_state: str, with_transition: str) -> monad.EitherMonad[str]:
    return (Pipe(state_map)
           .then(map_selector(extract_from_state_and_transition, (from_state, with_transition, None)))
           .then(list)
           .then(validate_and_extract_new_state)
           .flush())


@curry(3)
def map_selector(extractor_fn: Callable, map_tester: Tuple, state_map: StateTransitionMap) -> filter:
    return fn.select(fn.equality(extractor_fn, map_tester), state_map.stmap)

def validate_and_extract_new_state(state_map_items: List[Tuple[str, str, str]]) -> str:
    if fn.only_one(state_map_items):
        return monad.Right(state_map_items[0][2])
    return monad.Left("invalid transition")


def extract_from_state(state_map_item: Tuple) -> Tuple[str, None, None]:
    current, _transition, _new = state_map_item
    return (current, None, None)

def extract_from_state_and_transition(state_map_item: Tuple) -> Tuple:
    current, transition, _new = state_map_item
    return (current, transition, None)


def extract_transition(state_map_item: Tuple[str, str, str]) -> str:
    _current, transition, new = state_map_item
    return transition


def extract_transition_from_map(state_map: List[Tuple[str, str, str]]) -> str:
    return list(map(extract_transition, state_map))



import logging
from typing import Any

from aws_lambda_powertools import Logger, Tracer, Metrics
from metis_app import logger as base_logger
from metis_fn import singleton, fn


class Observer(singleton.Singleton):
    """
    Observer is a singleton which initialises Powertools Tracer, Logger, and Metrics.  Additional observables may
    also be added.  There is nothing special about the observer class, except that it constructs the necessary
    AWS Lambda Powertools classes.
    """
    logger = None
    tracer = None
    metric = None
    other_observers = {}

    def clear(self):
        self.logger = None
        self.tracer = None
        self.metrics = None

    def configure(self,
                  service_name: str,
                  metrics_namespace: str,
                  logger: Any = Logger,
                  tracer: Any = Tracer,
                  metrics: Any = Metrics,
                  custom_log_level: int = logging.INFO,
                  post_init_fn: callable = fn.identity,
                  **kwargs):
        """
        Creates a Lambda powertools logger, tracer, and metrics instance, which are then available via the
        Observer as a singleton.
        Args:
            service_name: The canonical name of the service to appear on logs and traces,
            metrics_namespace: The namespace of metrics
            logger: A logger cls, defaults to the powertools Logger
            tracer: A tracer cls, defaults to the powertools Tracer
            metrics: A metrics cls, defaults to the powertools Metrics
            custom_log_level: int log level, overrides the level set in the env or the LogConfig.
            post_init_fn: callable: Optionally provide a fn which will be called with self once initialised.
                          Use to add initialisation steps to the powertools instances.
            **kwargs:  any additional observers can be added to the observer singleton by providing them as the last
                       kwargs.  Then call the kwarg name on the observer singleton.  If it does not exist, an
                       AttributeError is raised.
        Returns:
            self
        >>>  obs = Observer().configure(service_name="a", metrics_namespace="b", my_observer=MyObserver())
        >>> obs.my_observer
        MyObserver()
        """
        self.service_name = service_name
        self.metrics_namespace = metrics_namespace
        self.logger = self._configure_logger(logger, custom_log_level)
        self.tracer = self._configure_tracer(tracer, self.service_name)
        self.metrics = self._configure_metrics(metrics, self.metrics_namespace, self.service_name)
        self.other_observers = kwargs
        post_init_fn(self)
        return self

    def add_custom(self, **kwargs):
        """
        Adds a custom observer to the Observer singleton.  It is available via the function named for the custom
        observer key in the kwargs
        """
        self.other_observers = kwargs
        return self

    def _configure_metrics(self, metrics_cls, namespace, service):
        if metrics_cls:
            return metrics_cls(namespace=self.metrics_namespace, service=self.service_name)
        return Metrics(namespace=self.metrics_namespace, service=self.service_name)

    def _configure_tracer(self, tracer_cls, service):
        if tracer_cls:
            return tracer_cls(service=self.service_name)
        return Tracer(service=self.service_name)

    def _configure_logger(self, logger_cls: Logger | Any, custom_log_level: int):
        """
        Sets the required logger in the LogConfig singleton.
        When the logger is the Powertools logger (the default), it must be wrapped in PowerToolsLoggerWrapper.
        Otherwise, passes the custom logger through.
        For the powertools logger, the level will be set from the env var POWERTOOLS_LOG_LEVEL
        """
        lgr = logger_cls(service=self.service_name)
        if isinstance(lgr, Logger):
            base_logger.LogConfig().configure(custom_logger=base_logger.PowerToolsLoggerWrapper(lgr))
        else:
            base_logger.LogConfig().configure(level=custom_log_level, custom_logger=lgr)
        return lgr

    def __getattr__(self, name):
        def obs(obj_name):
            return self.other_observers.get(obj_name, None)

        if name not in self.other_observers.keys():
            raise AttributeError
        return obs(name)

    @property
    def is_configured(self):
        return self.logger and self.tracer and self.metrics



import random

"""
Adds randomness to any retry concept, allowing competing processes to not (probably) compete for executing
a retry flow which updates shared state.

Used for the token expiry assessment, allowing several competing token gettings to:
1. Obtain a new token before it expires (end).
2. Apply a shrinking threshold to that end (width).
3. Assert a position in or out of the window (at)

Example (using times, rather than their Unix time values):
a. A token expiries at 12:00:00
b. The refresh of the token should occur before (or at) 11:00:00
c. The random refresh window started an hour before 11:00:00
d. Its now 10:05:00

> new.(00:60:00).in_window?(end: 11:00:00, at: 10:05:00)
=> possibly true
The at is in the window, therefore determine_in_window_chance(end, at) is called.
This picks a random number between 1 second .. 55 minutes (end - at)
If the random seconds are within the threshold for retry (THRESHOLD = 10 seconds) then the its in the window.
While its random that at is in the window, its approaches certainty the closer we get to the at, until
its completely certain (the at_right_end_of_window?(end, at) function)
"""

threshold = 10

def in_window(width, end, at):
    if not_in_window(width, end, at):
        return False
    if at_right_end_of_window(width, end, at):
        return True
    return determine_in_window_chance(width, end, at)

# This is the negative of in_window
def left_of_window(width, end, at):
    return not in_window(width=width, end=end, at=at)

def not_in_window(width, end, at):
    return (end - width) > at

def at_right_end_of_window(width, end, at):
    return ((end - at) == 0) or ((end - at) < threshold)

def determine_in_window_chance(width, end, at):
    chance = random.randint(1, int(end - at))
    return (width % chance) < threshold



from typing import Optional, Dict, Any, List, Callable
from datetime import datetime
from dataclasses import dataclass
from enum import Enum
from aws_lambda_powertools.utilities.data_classes import (
    S3Event,
    APIGatewayProxyEvent,
    KafkaEvent,
    EventBridgeEvent)

from . import tracer, error, app_serialisers, observable


class HttpStatusCode(Enum):
    OK = 200
    CREATED = 201
    BadRequest = 400
    Unauthorized = 401
    InternalServerError = 500


@dataclass
class DataClassAbstract:
    def replace(self, key, value):
        setattr(self, key, value)
        return self


@dataclass
class RequestEvent(DataClassAbstract):
    event: S3Event | APIGatewayProxyEvent | KafkaEvent | EventBridgeEvent
    kind: str
    request_function: Callable

    def returnable_session_state(self):
        return False


@dataclass
class S3Object(DataClassAbstract):
    bucket: str
    key: str
    event_time: Optional[datetime] = None
    object: Optional[list] = None
    meta: Optional[Dict] = None

    def s3_event_path(self):
        return "{bucket}/{key}".format(bucket=self.bucket, key=self.key)


@dataclass
class KafkaTopicEvent(DataClassAbstract):
    topic: str
    key: str | bytes
    value: str


@dataclass
class NoopEvent(RequestEvent):
    pass


@dataclass
class S3StateChangeEvent(RequestEvent):
    objects: List[S3Object]


@dataclass
class KafkaRecordsEvent(RequestEvent):
    events: List[KafkaTopicEvent]


@dataclass
class EventBridgePublishEvent(RequestEvent):
    topic: str
    body: dict


@dataclass
class ApiGatewayRequestEvent(RequestEvent):
    method: str
    headers: Dict
    path: str
    path_params: Dict
    body: str
    query_params: Optional[dict] = None
    web_session: Optional[Any] = None

    def clear_session(self):
        self.web_session = None
        self

    def returnable_session_state(self):
        return True


@dataclass
class Request(DataClassAbstract):
    event: RequestEvent
    context: dict
    tracer: tracer.Tracer
    observer: observable.Observer = None
    event_time: datetime = None
    app_request_context: Dict = None
    status_code: Optional[HttpStatusCode] = None
    request_handler: Optional[Callable] = None
    pip: Optional[dict] = None
    results: Optional[list] = None
    error: Optional[Any] = None
    response: Optional[dict] = None
    response_headers: Optional[dict] = None


class AppError(error.BaseError):
    @classmethod
    def dict_to_json_serialiser(cls):
        return app_serialisers.DictToJsonSerialiser

    def error(self):
        return type(self).dict_to_json_serialiser()(super().error())

    def as_dict(self):
        return super(type(self), self).error()

    def serialise(self):
        return self.error().serialise()



from typing import Tuple, Protocol, Union, Callable
from .crypto import jwk, jwt
from pymonad.tools import curry
from simple_memory_cache import GLOBAL_CACHE
import re

from metis_fn import monad, singleton, chronos
from . import http_adapter, http, circuit, error, crypto, cache

jwks_cache = GLOBAL_CACHE.MemoryCachedVar('jwks_cache')

event_authorisation_hdr_key = "Authorization"

JWKS = "JWKS"  # name in cache


class JwksGetError(error.BaseError):
    pass


class SubjectTokenConfig(singleton.Singleton):
    """
    The Subject Token config is setup by the library caller to provide 2 arguments:

    + jwks_endpoint. Returns the endpoint of the IDP for the JWKS call.
    + circuit_state_provider.  An implementation of the CircuitStateProviderProtocol
    + asserted_iss: The iss must be asserted when decoding the JWT.  Provide the iss str as it will appear in the JWT
    """

    def configure(self,
                  jwks_endpoint: str,
                  asserted_iss: str = "",
                  jwks_persistence_provider: cache.KeyValueCachePersistenceProviderProtocol = None,
                  circuit_state_provider: circuit.CircuitStateProviderProtocol = None):
        self.jwks_persistence_provider = jwks_persistence_provider
        self.jwks_endpoint = jwks_endpoint
        self.circuit_state_provider = circuit_state_provider
        self.asserted_iss = asserted_iss
        pass


def parse_generate_id_token(serialised_jwt):
    """
    Takes an encoded JWT and returns an verified id_token
    """
    if not serialised_jwt:
        return monad.Left(JwksGetError('No JWT found', code=401))
    return cacheable_jwks() >> crypto.decode_jwt(claims_to_assert(), serialised_jwt) >> crypto.to_id_token


def claims_to_assert():
    return {'iss': SubjectTokenConfig().asserted_iss,
            'exp': int(chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()]))}


def cacheable_jwks():
    return jwks_cache.get()


@jwks_cache.on_first_access
def _jwks():
    """
    Takes a key id (kid), requests the jwks from the identity jwks well-known service,
    returning a monadic jwt.JWKkeys
    """
    return monad.Right(jwks_resource()) >> jwks_from_cache >> get_jwks >> cache_jwks >> jwks_from_json


def jwk_cache_invalidate():
    jwks_cache.invalidate()

def cache_jwks(jwks: Tuple[int, str]):
    _status, keys = jwks
    if SubjectTokenConfig().jwks_persistence_provider:
        SubjectTokenConfig().jwks_persistence_provider.write(JWKS, keys)
    return monad.Right(jwks)

@monad.monadic_try(name="jwks_from_json")
def jwks_from_json(jwks: Tuple[int, str]):
    """
    Takes a JSON JWKS keyset and returns a wrapper object
    """
    _status, keys = jwks
    return jwk.JWKSet.from_json(keys)


@curry(3)
@monad.monadic_try(name="")
def rsa_key_from_kid(kid, op, jwks):
    """
    Takes a jwks object, extracts a public key based on KID and returns an
    openssl.rsa._RSAPublicKey
    """
    return jwks.get_key(kid).get_op_key(op)


def jwks_from_cache(endpoint):
    if not SubjectTokenConfig().jwks_persistence_provider:
        return monad.Right((endpoint, None))

    result = cache_reader(SubjectTokenConfig().jwks_persistence_provider)

    if result is None or result.is_left():
        return monad.Right((endpoint, None))
    return monad.Right((endpoint, result.value.value))


def cache_reader(provider: cache.KeyValueCachePersistenceProviderProtocol):
    if not hasattr(provider, 'read'):
        return None
    return provider.read(key=JWKS)


def get_jwks(endpoint_tuple: Tuple[str, Union[str, None]]):
    """
    Get the JWKS json from the Identity well known resource.
    Inject the http_status_exception to test for a Right from the request, but a failure HTTP status

    The endpoint tuple has the jwks_endpoint and either None or jwks obtained from cache
    """
    endpoint, jwks = endpoint_tuple
    if not jwks:
        return http_adapter.get(endpoint=endpoint,
                                name='jwks_service',
                                circuit_state_provider=SubjectTokenConfig().circuit_state_provider,
                                exception_test_fn=http.http_response_monad(__name__, http.extract_fn_raw),
                                error_cls=JwksGetError)

    return monad.Right((200, jwks))


def jwks_resource():
    return SubjectTokenConfig().jwks_endpoint


def parse_bearer_token(hdrs: dict) -> str | None:
    """
    Takes HTTP hdrs and extracts the bearer token from the Authorization header
    Support Camel and lowercase versions of the header
    """
    auth_hdr = hdrs.get(event_authorisation_hdr_key, hdrs.get(event_authorisation_hdr_key.lower()))
    if not auth_hdr:
        return None
    bearer_group = re.search(r'(Bearer)\s(.*)', auth_hdr)
    if not bearer_group or len(bearer_group.groups()) != 2:
        return None
    return bearer_group.group(2)


def bearer_token_hdr(jwt):
    return {'authorization': "Bearer {}".format(jwt)}
